'use strict';

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

const MongoOplog = require('mongo-oplog');
const MongoDB = require('mongodb');
const MongoDBClient = MongoDB.MongoClient;

const genericPool = require('generic-pool');

var _require = require('await-to-js');

const to = _require.to;


const ElasticSearchClient = require('./elasticSearchClient');
const QueueManager = require('../queue/queueManager');

const defaultOptions = {
  engine: 'mongodb',
  uri: 'mongodb://localhost:27017/test',
  max: 100,
  min: 1
};

class ConnectionManager {
  constructor(options) {
    this.models = {};

    const poolOptions = Object.assign({}, defaultOptions, {
      uri: options.uri,
      engine: options.engine
    });

    const uri = options.uri;


    const i = uri.indexOf('://');
    if (i < 0) {
      throw new Error('No scheme found in URI ' + uri);
    }

    // Initialize Oplog to perform action into ElasticSearch instance
    if (options.elasticSearch && options.elasticSearch.host) {
      if (this._elasticSearch && this._elasticSearch.oplog) this._elasticSearch.oplog.destroy();

      this._elasticSearch = new ElasticSearchClient(options.elasticSearch);
      this._queueManager = new QueueManager();

      const oplog = this._initializeMongoOplog(uri);

      Object.assign(this._elasticSearch, oplog);
    }

    this._pool = genericPool.createPool({
      create: () => new Promise((resolve, reject) => {
        //console.log(uri);
        MongoDBClient.connect(uri, _extends({
          poolSize: 1,
          useNewUrlParser: true
        }, options.connection), (err, client) => {
          if (err) reject(err);
          resolve(client);
        });
      }),
      destroy: client => new Promise(resolve => {
        client.close().then(() => {
          resolve();
        });
      })
    }, poolOptions);
  }

  /**
   * _parseUri(): Parse Mongo URI to get databasename
   * @param {*} uri
   */
  _parseUri(uri) {
    let opt = null;

    let i = uri.indexOf('?');
    if (i >= 0) {
      opt = uri.substring(i);
    }

    i = uri.indexOf('/', 10);

    return (uri.indexOf('/', 10) < 0 ? uri : uri.substring(0, i)) + '/local' + (opt || '');
  }

  /**
   * acquire(): Acquire a resource from the DBClient
   *
   * @returns {Promise}
   */
  acquire() {
    var _this = this;

    return _asyncToGenerator(function* () {
      if (!_this._pool) throw new Error('You must first connect to the DB before loading/saving documents.');

      try {
        return yield _this._pool.acquire();
      } catch (error) {
        return error;
      }
    })();
  }

  /**
   * acquire(): Release a borrowed resource from the DBClient
   *
   * @returns {Promise}
   */
  release(conn) {
    var _this2 = this;

    return _asyncToGenerator(function* () {
      if (!_this2._pool) throw new Error('You must first connect to the DB before loading/saving documents.');
      try {
        return yield _this2._pool.release(conn);
      } catch (error) {
        return error;
      }
    })();
  }

  /**
   * Close current client
   *
   * @returns {Promise}
   */
  close() {
    var _this3 = this;

    return _asyncToGenerator(function* () {
      if (!_this3._pool) throw new Error('There is no connection currently active.');
      try {
        return yield _this3._pool.drain();
      } catch (error) {
        return error;
      }
    })();
  }

  /**
   * _initializeMongoOplog(): Initialize all listeners for
   * Mongo oplog to perform operations into ES instance
   *
   * @param {String} uri
   */
  _initializeMongoOplog(uri) {
    var _this4 = this;

    return _asyncToGenerator(function* () {
      const oplogUri = _this4._parseUri(uri);

      const oplog = MongoOplog(oplogUri);

      yield oplog.tail();

      /*oplog.on('op', data => {
        console.log('OP', data);
      });*/

      oplog.on('insert', function (doc) {
        const model = _this4.models[doc.ns.split('.')[1]];
        if (model && model._schemaOptions && model._schemaOptions.elasticSearchIndexes) {
          _this4._queueManager.pushCommand(_this4._insertDocOnES.bind(_this4, doc), 'insert', function (err, result) {
            if (err) throw err;
          });
        }
      });

      oplog.on('update', function (doc) {
        const model = _this4.models[doc.ns.split('.')[1]];
        if (model && model._schemaOptions && model._schemaOptions.elasticSearchIndexes) {
          _this4._queueManager.pushCommand(_this4._updateDocOnES.bind(_this4, doc), 'update', function (err, result) {
            if (err) throw err;
          });
        }
      });

      oplog.on('delete', function (doc) {
        const model = _this4.models[doc.ns.split('.')[1]];
        if (model && model._schemaOptions && model._schemaOptions.elasticSearchIndexes) {
          _this4._queueManager.pushCommand(_this4._deleteDocOnES.bind(_this4, doc), 'delete', function (err, result) {
            if (err) throw err;
          });
        }
      });

      /*oplog.on('error', error => {
        console.log('ERROR', error);
      });*/

      oplog.on('end', _asyncToGenerator(function* () {
        yield oplog.destroy();
      }));

      return oplog;
    })();
  }

  /**
   * _insertDocOnES(): Insert doc abstract into ES
   *
   * @param {Object} doc
   */
  _insertDocOnES(doc) {
    var _this5 = this;

    return _asyncToGenerator(function* () {
      const model = _this5.models[doc.ns.split('.')[1]];
      const elasticSearchIndexes = model._schemaOptions.elasticSearchIndexes;


      let payload = null;
      Object.keys(elasticSearchIndexes).forEach(function (indexKey) {
        if (Object.keys(doc.o).includes(indexKey)) {
          payload = _extends({}, payload, {
            [indexKey]: doc.o[indexKey]
          });
        }
      });

      if (payload) {
        var _ref2 = yield _this5._createIndex(elasticSearchIndexes, doc),
            _ref3 = _slicedToArray(_ref2, 2);

        const index = _ref3[0],
              collection = _ref3[1];


        if (model._schemaOptions && model._schemaOptions.inheritOptions) {
          const discriminatorKey = model._schemaOptions.inheritOptions.discriminatorKey;

          payload = _extends({
            collection,
            [discriminatorKey]: doc.o[discriminatorKey]
          }, payload);
        } else {
          payload = _extends({
            collection
          }, payload);
        }

        yield _this5._elasticSearch.addDocument(index, 'doc', doc.o._id.toString(), payload);
      }
    })();
  }

  /**
   * _updateDocOnES(): Update ES record
   *
   * @param {Object} doc
   */
  _updateDocOnES(doc) {
    var _this6 = this;

    return _asyncToGenerator(function* () {
      const elasticSearchIndexes = _this6.models[doc.ns.split('.')[1]]._schemaOptions.elasticSearchIndexes;

      let payload = null;
      Object.keys(elasticSearchIndexes).forEach(function (indexKey) {
        Object.keys(doc.o).forEach(function (updateOperator) {
          if (Object.keys(doc.o[updateOperator]).includes(indexKey)) {
            payload = _extends({}, payload, {
              [indexKey]: doc.o[updateOperator][indexKey]
            });
          }
        });
      });

      if (payload) {
        var _ref4 = yield _this6._createIndex(elasticSearchIndexes, doc),
            _ref5 = _slicedToArray(_ref4, 2);

        const index = _ref5[0],
              collection = _ref5[1];


        yield _this6._elasticSearch.updateDocument(index, 'doc', doc.o2._id.toString(), payload);
      }
    })();
  }

  /**
   * _deleteDocOnES(): Delete ES record
   *
   * @param {Object} doc
   */
  _deleteDocOnES(doc) {
    var _this7 = this;

    return _asyncToGenerator(function* () {
      var _doc$ns$split = doc.ns.split('.'),
          _doc$ns$split2 = _slicedToArray(_doc$ns$split, 2);

      const index = _doc$ns$split2[0],
            collection = _doc$ns$split2[1];

      yield _this7._elasticSearch.deleteDocument(index, 'doc', doc.o._id.toString());
    })();
  }

  /**
   * _createIndex(): Check if an index exists, if not
   * create it
   *
   * @param {Object} indexes
   * @param {Object} doc
   *
   * @returns {Array} [index, collection]
   */
  _createIndex(indexes, doc) {
    var _this8 = this;

    return _asyncToGenerator(function* () {
      var _doc$ns$split3 = doc.ns.split('.'),
          _doc$ns$split4 = _slicedToArray(_doc$ns$split3, 2);

      const index = _doc$ns$split4[0],
            collection = _doc$ns$split4[1];

      var _ref6 = yield to(_this8._elasticSearch.indexExists(index)),
          _ref7 = _slicedToArray(_ref6, 2);

      const error = _ref7[0],
            indexExists = _ref7[1];


      if (!indexExists) {
        var _ref8 = yield to(_this8._elasticSearch.createIndex(index)),
            _ref9 = _slicedToArray(_ref8, 2);

        const error = _ref9[0],
              result = _ref9[1];

        /*Object.keys(indexes).forEach(index => {
          indexes = {
            ...indexes,
            [index]: {
              ...indexes[index],
              store: true,
            },
          };
        });
          await this._elasticSearch.putMapping(index, collection, indexes);*/
      }

      return [index, collection];
    })();
  }

  // ===============================
  // PUBLIC ELASTIC SEARCH API =====
  // ===============================

  /**
   * searchOnES(): Get documents from ES based
   * on its query
   *
   * @param {String} index
   * @param {String} collection
   * @param {Object} query
   *
   * @returns {Array}
   */
  searchOnES(index, collection, query = []) {
    var _this9 = this;

    return _asyncToGenerator(function* () {
      const model = _this9.models[collection];
      let match = {};

      if (model._schemaOptions && model._schemaOptions.inheritOptions) {
        const discriminatorKey = model._schemaOptions.inheritOptions.discriminatorKey;

        match = [{
          match: {
            collection: model._modelName
          }
        }, {
          match: {
            [discriminatorKey]: collection
          }
        }];
      } else {
        match = [{
          match: {
            collection: model._modelName
          }
        }];
      }

      query = {
        bool: {
          must: [...match, ...query]
        }
      };

      return yield _this9._elasticSearch.search(index, 'doc', query);
    })();
  }

  /**
   * deleteIndexOnES(): Delete index from ES
   *
   * @param {String} index
   * @param {String} collection
   * @param {Object} query
   *
   * @returns {Array}
   */
  deleteIndexOnES(index) {
    var _this10 = this;

    return _asyncToGenerator(function* () {
      return yield _this10._elasticSearch.deleteIndex(index);
    })();
  }
}

module.exports = ConnectionManager;