'use strict';

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

//- A Promise-based middleware implementation

let nextTick = 'undefined' !== typeof process ? process.nextTick : 'undefined' !== typeof setImmediate ? setImmediate : setTimeout;

class Middleware {
  /** create a new Middleware */
  constructor() {
    this.middlewares = [];
  }

  /** add a function to middleware stack */
  use(fn) {
    if (!fn || typeof fn !== 'function') {
      throw new TypeError("Argument to use must be of type 'Function'");
    }

    this.middlewares.push(fn);
  }

  /** execute the middlewares */
  exec(...args) {
    var _this = this;

    return new Promise((resolve, reject) => {
      // check for early exit
      if (!this.middlewares.length) return resolve(args);

      // kickstart the chain
      let _execute = (() => {
        var _ref = _asyncToGenerator(function* (i, ...args0) {
          try {
            yield _this.middlewares[i](function (...returnValue) {
              if (returnValue.length && returnValue[0] instanceof Error) return returnValue[0];else if (i >= _this.middlewares.length - 1) return returnValue;else return _execute(i + 1, ...returnValue);
            }, ...args0);
          } catch (err) {
            return reject(err);
          }
        });

        return function _execute(_x) {
          return _ref.apply(this, arguments);
        };
      })();

      return resolve(_execute(0, ...args));
    });
  }
}

module.exports = Middleware;